from __future__ import annotations

from collections.abc import Iterator
from dataclasses import replace
from typing import Any

from omnichunk.context.entities import enrich_parent_links, extract_entities
from omnichunk.context.format import format_contextualized_text
from omnichunk.context.imports import build_import_infos, filter_imports_for_chunk
from omnichunk.context.scope import build_scope_tree, find_scope_chain
from omnichunk.context.siblings import SiblingIndex, build_sibling_index, detect_siblings_for_chunk
from omnichunk.parser.tree_sitter import parse_code
from omnichunk.sizing.counter import make_token_counter
from omnichunk.sizing.nws import get_nws_count, preprocess_nws_cumsum
from omnichunk.types import (
    ByteRange,
    Chunk,
    ChunkContext,
    ChunkOptions,
    ContentType,
    EntityInfo,
    LineRange,
)
from omnichunk.util.detect import detect_language
from omnichunk.util.text_index import TextIndex
from omnichunk.windowing.greedy import assign_windows_for_nodes, assign_windows_for_ranges
from omnichunk.windowing.merge import merge_adjacent_windows
from omnichunk.windowing.overlap import apply_token_overlap, build_line_overlap_text


class CodeEngine:
    def chunk(self, filepath: str, content: str, options: ChunkOptions) -> list[Chunk]:
        chunks = list(self._iter_base_chunks(filepath, content, options))
        chunks = _finalize_chunk_indexes(chunks)

        if options.overlap is not None and chunks:
            chunks = apply_token_overlap(
                chunks,
                overlap=options.overlap,
                max_chunk_size=max(1, options.max_chunk_size),
            )

        return _finalize_chunk_indexes(chunks)

    def stream(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        """Yield base chunks lazily. ``total_chunks`` is always ``-1``.

        Token overlap (``options.overlap``) is not applied in streaming mode; use
        :meth:`chunk` when overlap is required.
        """
        for idx, chunk in enumerate(self._iter_base_chunks(filepath, content, options)):
            yield _with_unknown_total(chunk, idx)

    def _iter_base_chunks(
        self,
        filepath: str,
        content: str,
        options: ChunkOptions,
    ) -> Iterator[Chunk]:
        if not content.strip():
            return

        language = options.language or detect_language(filepath=filepath, content=content)
        parse_result = parse_code(content, language, filepath=filepath)

        entities = enrich_parent_links(extract_entities(content, language, parse_result.tree))
        scope_tree = build_scope_tree(entities)
        import_infos = build_import_infos(entities)
        sibling_index = build_sibling_index(scope_tree.all_entities)

        precomputed_text_index = options._precomputed_text_index
        if isinstance(precomputed_text_index, TextIndex):
            text_index = precomputed_text_index
        else:
            text_index = TextIndex(content)

        cumsum = options._precomputed_nws_cumsum
        if cumsum is None:
            cumsum = preprocess_nws_cumsum(content, backend=options.nws_backend)

        raw_bytes = text_index.raw_bytes

        max_size_nws = _to_internal_nws_limit(options.max_chunk_size, options.size_unit)
        min_size_nws = max(1, _to_internal_nws_limit(options.min_chunk_size, options.size_unit))

        ast_nodes = _collect_ast_root_nodes(parse_result.tree)
        if ast_nodes:
            windows = assign_windows_for_nodes(
                ast_nodes,
                cumsum=cumsum,
                max_size=max_size_nws,
                code=content,
            )
        else:
            candidate_ranges = _collect_candidate_ranges(
                entities, parse_result.tree, len(raw_bytes)
            )
            windows = assign_windows_for_ranges(
                candidate_ranges,
                cumsum=cumsum,
                max_size=max_size_nws,
                code=content,
            )

        merged_windows = merge_adjacent_windows(
            windows,
            max_size=max_size_nws,
            min_size=min_size_nws,
        )

        chunk_ranges = _windows_to_contiguous_ranges(merged_windows, len(raw_bytes))
        chunk_ranges = _merge_whitespace_only_ranges(chunk_ranges, raw_bytes)
        token_counter = make_token_counter(options.tokenizer, chunk_size=options.max_chunk_size)

        previous_text = ""
        chunk_index = 0

        for start, end in chunk_ranges:
            if end <= start:
                continue

            text = raw_bytes[start:end].decode("utf-8", errors="replace")
            if not text.strip():
                continue

            byte_range = ByteRange(start=start, end=end)
            line_range = LineRange(
                start=text_index.line_for_byte(start),
                end=text_index.line_for_byte(max(start, end - 1)),
            )

            context = _build_context(
                filepath=filepath,
                language=language,
                chunk_range=byte_range,
                entities=entities,
                scope_tree=scope_tree,
                sibling_index=sibling_index,
                import_infos=import_infos,
                options=options,
                parse_errors=parse_result.errors,
            )

            overlap_text = build_line_overlap_text(previous_text, options.overlap_lines)
            contextualized_text = (
                text
                if options.context_mode == "none"
                else format_contextualized_text(text, context, overlap_text=overlap_text)
            )

            yield Chunk(
                text=text,
                contextualized_text=contextualized_text,
                byte_range=byte_range,
                line_range=line_range,
                index=chunk_index,
                total_chunks=-1,
                context=context,
                token_count=token_counter(text),
                char_count=len(text),
                nws_count=get_nws_count(cumsum, start, end),
            )

            previous_text = text
            chunk_index += 1


def _build_context(
    *,
    filepath: str,
    language: str,
    chunk_range: ByteRange,
    entities: list[EntityInfo],
    scope_tree: Any,
    sibling_index: SiblingIndex,
    import_infos: list[Any],
    options: ChunkOptions,
    parse_errors: list[str],
) -> ChunkContext:
    if options.context_mode == "none":
        return ChunkContext(
            filepath=filepath,
            language=language,  # type: ignore[arg-type]
            content_type=ContentType.CODE,
            parse_errors=list(parse_errors),
        )

    overlapping_entities = _entities_in_range(entities, chunk_range)
    scope = find_scope_chain(scope_tree, chunk_range)
    breadcrumb = [e.name for e in reversed(scope) if e.name]

    siblings = []
    if options.context_mode == "full" and options.sibling_detail != "none":
        siblings = detect_siblings_for_chunk(
            sibling_index,
            chunk_range,
            max_siblings=max(0, options.max_siblings),
        )
        if options.sibling_detail == "names":
            siblings = [replace(s, signature="") for s in siblings]

    imports = []
    if options.include_imports:
        if options.filter_imports:
            imports = filter_imports_for_chunk(
                import_infos, [e.signature for e in overlapping_entities if e.signature]
            )
        else:
            imports = import_infos

    if options.context_mode == "minimal":
        return ChunkContext(
            filepath=filepath,
            language=language,  # type: ignore[arg-type]
            content_type=ContentType.CODE,
            breadcrumb=breadcrumb,
            entities=[],
            siblings=[],
            imports=imports[:5],
            parse_errors=list(parse_errors),
        )

    return ChunkContext(
        filepath=filepath,
        language=language,  # type: ignore[arg-type]
        content_type=ContentType.CODE,
        scope=scope,
        breadcrumb=breadcrumb,
        entities=overlapping_entities,
        siblings=siblings,
        imports=imports,
        parse_errors=list(parse_errors),
    )


def _entities_in_range(entities: list[EntityInfo], chunk_range: ByteRange) -> list[EntityInfo]:
    out: list[EntityInfo] = []
    for entity in entities:
        br = entity.byte_range
        if br is None:
            continue
        if br.end <= chunk_range.start or br.start >= chunk_range.end:
            continue
        is_partial = br.start < chunk_range.start or br.end > chunk_range.end
        if is_partial:
            out.append(replace(entity, is_partial=True))
        else:
            out.append(entity)
    return out


def _collect_candidate_ranges(
    entities: list[EntityInfo],
    tree: Any | None,
    total_bytes: int,
) -> list[tuple[int, int]]:
    ranges: list[tuple[int, int]] = []

    for entity in entities:
        if entity.byte_range is None:
            continue
        if entity.type.value in {"import", "export"}:
            continue
        start = entity.byte_range.start
        end = entity.byte_range.end
        if end > start:
            ranges.append((start, end))

    if not ranges and tree is not None:
        root = getattr(tree, "root_node", None)
        if root is not None:
            for child in getattr(root, "children", []) or []:
                start = int(getattr(child, "start_byte", 0))
                end = int(getattr(child, "end_byte", 0))
                if end > start:
                    ranges.append((start, end))

    if not ranges and total_bytes > 0:
        ranges.append((0, total_bytes))

    ranges.sort(key=lambda r: (r[0], r[1]))

    merged: list[tuple[int, int]] = []
    for start, end in ranges:
        start = max(0, min(start, total_bytes))
        end = max(0, min(end, total_bytes))
        if end <= start:
            continue
        if not merged or start > merged[-1][1]:
            merged.append((start, end))
        else:
            merged[-1] = (merged[-1][0], max(merged[-1][1], end))

    return merged


def _collect_ast_root_nodes(tree: Any | None) -> list[Any]:
    if tree is None:
        return []
    root = getattr(tree, "root_node", None)
    if root is None:
        return []

    candidates = list(getattr(root, "named_children", []) or [])
    if not candidates:
        candidates = list(getattr(root, "children", []) or [])

    out: list[Any] = []
    for node in candidates:
        start = int(getattr(node, "start_byte", 0))
        end = int(getattr(node, "end_byte", 0))
        if end > start:
            out.append(node)
    return out


def _windows_to_contiguous_ranges(
    windows: list[list[Any]], total_bytes: int
) -> list[tuple[int, int]]:
    if total_bytes <= 0:
        return []
    if not windows:
        return [(0, total_bytes)]

    sorted_windows = sorted(windows, key=lambda w: min(i.start for i in w))
    cut_points: list[int] = []
    for window in sorted_windows:
        end = max(item.end for item in window)
        cut_points.append(end)

    ranges: list[tuple[int, int]] = []
    cursor = 0
    for cut in cut_points:
        cut = max(cursor, min(cut, total_bytes))
        if cut > cursor:
            ranges.append((cursor, cut))
            cursor = cut
    if cursor < total_bytes:
        ranges.append((cursor, total_bytes))

    return ranges


def _merge_whitespace_only_ranges(
    ranges: list[tuple[int, int]],
    raw_bytes: bytes,
) -> list[tuple[int, int]]:
    if not ranges:
        return []

    merged: list[tuple[int, int]] = []
    pending_prefix: tuple[int, int] | None = None

    for start, end in ranges:
        segment = raw_bytes[start:end].decode("utf-8", errors="replace")
        if segment.strip():
            if pending_prefix is not None:
                start = pending_prefix[0]
                pending_prefix = None
            merged.append((start, end))
            continue

        if merged:
            prev_start, prev_end = merged[-1]
            merged[-1] = (prev_start, end)
        else:
            pending_prefix = (start, end)

    if pending_prefix is not None and merged:
        first_start, first_end = merged[0]
        merged[0] = (pending_prefix[0], first_end)

    if not merged and ranges:
        return [ranges[0]]

    return merged


def _to_internal_nws_limit(size: int, unit: str) -> int:
    size = max(1, int(size))
    if unit == "tokens":
        return size * 4
    if unit == "chars":
        return size
    return size


def _finalize_chunk_indexes(chunks: list[Chunk]) -> list[Chunk]:
    total = len(chunks)
    return [
        Chunk(
            text=chunk.text,
            contextualized_text=chunk.contextualized_text,
            byte_range=chunk.byte_range,
            line_range=chunk.line_range,
            index=idx,
            total_chunks=total,
            context=chunk.context,
            token_count=chunk.token_count,
            char_count=chunk.char_count,
            nws_count=chunk.nws_count,
        )
        for idx, chunk in enumerate(chunks)
    ]


def _with_unknown_total(chunk: Chunk, index: int) -> Chunk:
    return Chunk(
        text=chunk.text,
        contextualized_text=chunk.contextualized_text,
        byte_range=chunk.byte_range,
        line_range=chunk.line_range,
        index=index,
        total_chunks=-1,
        context=chunk.context,
        token_count=chunk.token_count,
        char_count=chunk.char_count,
        nws_count=chunk.nws_count,
    )



"""Structure-aware chunking; :class:`Chunker` adds ``achunk``, ``astream``, and ``abatch``."""

from . import plugins
from .budget import BudgetResult, TokenBudgetOptimizer
from .chunker import Chunker, chunk, chunk_directory, chunk_file, hierarchical_chunk
from .dedup import dedup_chunks
from .diff import chunk_diff
from .eval import ChunkEvalScores, EvalReport, eval_report_to_dict, evaluate_chunks
from .formats import (
    FormatSegment,
    LoadedDocument,
    chunk_loaded_document,
    load_docx_bytes,
    load_ipynb,
    load_latex,
    load_pdf_bytes,
)
from .graph import ChunkEdge, ChunkGraph, EntityNode, build_chunk_graph
from .otel import maybe_span
from .plugins import (
    list_registered_formatters,
    list_registered_parsers,
    register_formatter,
    register_parser,
)
from .propositions import Proposition
from .semantic import (
    SemanticBoundaryResult,
    SemanticSplitter,
    build_tfidf_matrix,
    detect_semantic_boundaries,
    detect_topic_shifts,
    split_sentences,
)
from .serialization import (
    chunk_from_dict,
    chunks_to_pinecone_vectors,
    chunks_to_supabase_rows,
    chunks_to_weaviate_objects,
    stable_chunk_id,
)
from .store import ChunkStore, SyncResult
from .types import (
    BatchResult,
    ByteRange,
    Chunk,
    ChunkContext,
    ChunkDiff,
    ChunkNode,
    ChunkOptions,
    ChunkQualityScore,
    ChunkStats,
    ChunkTree,
    ContentType,
    EntityInfo,
    EntityType,
    ImportInfo,
    LineRange,
    SiblingInfo,
    UpsertBatch,
)

__version__ = "0.10.1"

__all__ = [
    "BatchResult",
    "BudgetResult",
    "ByteRange",
    "Chunk",
    "ChunkContext",
    "ChunkEvalScores",
    "ChunkDiff",
    "ChunkEdge",
    "ChunkGraph",
    "ChunkNode",
    "ChunkOptions",
    "ChunkQualityScore",
    "ChunkStats",
    "ChunkTree",
    "ChunkStore",
    "Chunker",
    "ContentType",
    "EvalReport",
    "FormatSegment",
    "LoadedDocument",
    "EntityInfo",
    "EntityNode",
    "EntityType",
    "ImportInfo",
    "LineRange",
    "Proposition",
    "SemanticBoundaryResult",
    "SemanticSplitter",
    "SiblingInfo",
    "SyncResult",
    "TokenBudgetOptimizer",
    "UpsertBatch",
    "__version__",
    "build_chunk_graph",
    "build_tfidf_matrix",
    "chunk",
    "chunk_diff",
    "chunk_from_dict",
    "chunk_loaded_document",
    "dedup_chunks",
    "evaluate_chunks",
    "eval_report_to_dict",
    "chunk_directory",
    "chunk_file",
    "detect_semantic_boundaries",
    "detect_topic_shifts",
    "hierarchical_chunk",
    "load_docx_bytes",
    "load_ipynb",
    "load_latex",
    "load_pdf_bytes",
    "split_sentences",
    "stable_chunk_id",
    "maybe_span",
    "chunks_to_pinecone_vectors",
    "chunks_to_supabase_rows",
    "chunks_to_weaviate_objects",
    "list_registered_formatters",
    "list_registered_parsers",
    "plugins",
    "register_formatter",
    "register_parser",
]



"""Tests for Rust cosine similarity backend — skipped if Rust not built."""

from __future__ import annotations

import numpy as np
import pytest

from omnichunk.semantic.boundaries import _batch_cosine_similarity
from omnichunk.sizing.rust_accel import batch_cosine_similarity_adjacent_rust


def test_rust_cosine_matches_numpy_if_available() -> None:
    rng = np.random.default_rng(42)
    emb = rng.standard_normal((10, 16)).astype(np.float32)
    rust_result = batch_cosine_similarity_adjacent_rust(emb)
    if rust_result is None:
        pytest.skip("Rust backend not available")
    norms = np.linalg.norm(emb, axis=1, keepdims=True)
    norms = np.where(norms == 0, 1.0, norms)
    normalized = emb / norms
    expected = np.einsum("id,id->i", normalized[:-1], normalized[1:])
    np.testing.assert_allclose(rust_result, expected, atol=1e-5)


def test_batch_cosine_similarity_output_shape() -> None:
    emb = np.eye(5, 8)
    result = _batch_cosine_similarity(emb)
    assert result.shape == (4,)


def test_batch_cosine_similarity_identical_vectors() -> None:
    v = np.array([[1.0, 0.0, 0.0]] * 4)
    result = _batch_cosine_similarity(v)
    np.testing.assert_allclose(result, 1.0, atol=1e-6)


def test_batch_cosine_similarity_zero_vectors() -> None:
    v = np.zeros((3, 4))
    result = _batch_cosine_similarity(v)
    assert result.shape == (2,)
    assert not np.any(np.isnan(result))


def test_batch_cosine_similarity_orthogonal() -> None:
    v = np.eye(4, 4)
    result = _batch_cosine_similarity(v)
    np.testing.assert_allclose(result, 0.0, atol=1e-6)



from __future__ import annotations

from collections.abc import Sequence

import numpy as np

from omnichunk.sizing.rust_accel import (
    NwsBackendStatus,
    maybe_preprocess_nws_cumsum_rust,
    nws_backend_status,
)

_WHITESPACE_BYTES = np.array([9, 10, 11, 12, 13, 32], dtype=np.uint8)
_WHITESPACE_LOOKUP = np.zeros(256, dtype=np.bool_)
_WHITESPACE_LOOKUP[_WHITESPACE_BYTES] = True


def preprocess_nws_cumsum(code: str, *, backend: str | None = None) -> np.ndarray:
    """Build cumulative non-whitespace byte counts for O(1) range queries."""
    rust_result = maybe_preprocess_nws_cumsum_rust(code, backend=backend)
    if rust_result is not None:
        return rust_result

    return preprocess_nws_cumsum_python(code)


def preprocess_nws_cumsum_python(code: str) -> np.ndarray:
    """Pure-Python/Numpy NWS preprocessing used when Rust backend is unavailable."""
    if not code:
        return np.array([0], dtype=np.int64)

    raw = np.frombuffer(code.encode("utf-8"), dtype=np.uint8)
    mask = ~_WHITESPACE_LOOKUP[raw]
    cumsum = np.empty(mask.shape[0] + 1, dtype=np.int64)
    cumsum[0] = 0
    np.cumsum(mask, out=cumsum[1:])
    return cumsum


def get_nws_backend_status(backend: str | None = None) -> NwsBackendStatus:
    """Return active NWS backend details for profiling and diagnostics."""
    return nws_backend_status(backend)


def get_nws_count(cumsum: Sequence[int] | np.ndarray, start: int, end: int) -> int:
    """Return non-whitespace count for [start, end) byte range."""
    if start < 0:
        start = 0
    max_index = len(cumsum) - 1
    if end > max_index:
        end = max_index
    if end <= start:
        return 0
    return int(cumsum[end]) - int(cumsum[start])



"""
Hierarchical chunk tree builder.

Level 0: chunk at ``levels[0]``. Each coarser level groups either consecutive
leaves (level 1) or consecutive nodes at the previous level (level 2+), so
parent byte ranges exactly span their children.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import replace

from omnichunk.chunker import Chunker
from omnichunk.types import (
    ByteRange,
    Chunk,
    ChunkContext,
    ChunkNode,
    ChunkOptions,
    ChunkTree,
    EntityInfo,
    LineRange,
)


class _MutableNode:
    __slots__ = ("chunk", "level", "parent", "children", "leaf_lo", "leaf_hi")

    def __init__(
        self,
        chunk: Chunk,
        level: int,
        parent: int | None,
        children: list[int],
        leaf_lo: int,
        leaf_hi: int,
    ) -> None:
        self.chunk = chunk
        self.level = level
        self.parent = parent
        self.children = children
        self.leaf_lo = leaf_lo
        self.leaf_hi = leaf_hi


def _validate_levels(levels: Sequence[int]) -> tuple[int, ...]:
    lv = tuple(int(x) for x in levels)
    if len(lv) < 2:
        raise ValueError("hierarchical_chunk requires at least 2 levels")
    if any(x <= 0 for x in lv):
        raise ValueError("all levels must be > 0")
    if any(lv[i] > lv[i + 1] for i in range(len(lv) - 1)):
        raise ValueError("levels must be sorted in ascending order")
    return lv


def _chunk_at_level(
    filepath: str,
    content: str,
    max_size: int,
    size_unit: str,
    tokenizer: object,
    chunker_options: dict[str, object],
) -> list[Chunk]:
    opts = {k: v for k, v in chunker_options.items() if k in ChunkOptions.__dataclass_fields__}
    chunker = Chunker(
        max_chunk_size=max_size,
        min_chunk_size=max(1, max_size // 4),
        size_unit=size_unit,
        tokenizer=tokenizer,
        **opts,
    )
    return chunker.chunk(filepath, content)


def _merge_parent_context(children: list[Chunk], filepath: str) -> ChunkContext:
    first = children[0].context
    seen: dict[tuple[str, str], EntityInfo] = {}
    crumbs: list[str] = []
    for ch in children:
        crumbs.extend(ch.context.breadcrumb)
        for ent in ch.context.entities:
            seen[(ent.name, ent.type.value)] = ent
    return ChunkContext(
        filepath=filepath,
        language=first.language,
        content_type=first.content_type,
        scope=list(first.scope),
        breadcrumb=crumbs,
        entities=list(seen.values()),
        siblings=[],
        imports=list(first.imports) if first.imports else [],
        heading_hierarchy=list(first.heading_hierarchy),
        section_type=first.section_type,
        parse_errors=list(first.parse_errors),
    )


def _build_parent_chunk_from_leaf_slice(
    leaves: list[Chunk],
    leaf_lo: int,
    leaf_hi: int,
    content: str,
    filepath: str,
) -> Chunk:
    first_leaf = leaves[leaf_lo]
    last_leaf = leaves[leaf_hi]
    child_slice = leaves[leaf_lo : leaf_hi + 1]
    byte_start = first_leaf.byte_range.start
    byte_end = last_leaf.byte_range.end
    raw = content.encode("utf-8")
    text = raw[byte_start:byte_end].decode("utf-8")
    line_start = first_leaf.line_range.start
    line_end = last_leaf.line_range.end
    context = _merge_parent_context(child_slice, filepath)
    return Chunk(
        text=text,
        contextualized_text=text,
        byte_range=ByteRange(byte_start, byte_end),
        line_range=LineRange(line_start, max(line_start, line_end)),
        index=0,
        total_chunks=-1,
        context=context,
        token_count=sum(c.token_count for c in child_slice),
        char_count=sum(c.char_count for c in child_slice),
        nws_count=sum(c.nws_count for c in child_slice),
    )


def _leaf_metric(leaf: Chunk, size_unit: str) -> int:
    if size_unit == "chars":
        v = leaf.char_count
    elif size_unit == "tokens":
        v = leaf.token_count
    else:
        v = leaf.nws_count
    return max(1, int(v))


def _group_leaves_for_level(
    leaves: list[Chunk],
    max_size: int,
    size_unit: str,
    _tokenizer: object,
) -> list[list[int]]:
    groups: list[list[int]] = []
    current: list[int] = []
    current_size = 0
    for i, leaf in enumerate(leaves):
        leaf_size = _leaf_metric(leaf, size_unit)
        if current and current_size + leaf_size > max_size:
            groups.append(current)
            current = [i]
            current_size = leaf_size
        else:
            current.append(i)
            current_size += leaf_size
    if current:
        groups.append(current)
    return groups


def _node_metric(mnodes: list[_MutableNode], gi: int, size_unit: str) -> int:
    ch = mnodes[gi].chunk
    if size_unit == "chars":
        v = ch.char_count
    elif size_unit == "tokens":
        v = ch.token_count
    else:
        v = ch.nws_count
    return max(1, int(v))


def _group_prev_level_nodes(
    mnodes: list[_MutableNode],
    prev_indices: list[int],
    max_size: int,
    size_unit: str,
) -> list[list[int]]:
    groups: list[list[int]] = []
    current: list[int] = []
    current_size = 0
    for gi in prev_indices:
        sz = _node_metric(mnodes, gi, size_unit)
        if current and current_size + sz > max_size:
            groups.append(current)
            current = [gi]
            current_size = sz
        else:
            current.append(gi)
            current_size += sz
    if current:
        groups.append(current)
    return groups


def _finalize_chunk_indices(mnodes: list[_MutableNode]) -> None:
    max_level = max(m.level for m in mnodes) if mnodes else -1
    for level in range(max_level + 1):
        idxs = [i for i, m in enumerate(mnodes) if m.level == level]
        tot = len(idxs)
        for k, mi in enumerate(idxs):
            m = mnodes[mi]
            m.chunk = replace(m.chunk, index=k, total_chunks=tot)


def build_chunk_tree(
    filepath: str,
    content: str,
    *,
    levels: Sequence[int],
    size_unit: str = "chars",
    tokenizer: object = None,
    **chunker_options: object,
) -> ChunkTree:
    lv = _validate_levels(levels)
    L = len(lv)
    opts = dict(chunker_options)
    leaves = _chunk_at_level(filepath, content, lv[0], size_unit, tokenizer, opts)
    if not leaves:
        return ChunkTree(nodes=[], level_count=L)

    mnodes: list[_MutableNode] = []
    for i, leaf in enumerate(leaves):
        mnodes.append(
            _MutableNode(
                chunk=leaf,
                level=0,
                parent=None,
                children=[],
                leaf_lo=i,
                leaf_hi=i,
            )
        )

    prev_level_indices = list(range(len(mnodes)))

    for ell in range(1, L):
        max_sz = lv[ell]
        if ell == 1:
            groups = _group_leaves_for_level(leaves, max_sz, size_unit, tokenizer)
        else:
            groups = _group_prev_level_nodes(mnodes, prev_level_indices, max_sz, size_unit)

        new_level_indices: list[int] = []
        for group in groups:
            if ell == 1:
                child_global = tuple(sorted(prev_level_indices[j] for j in group))
                lo, hi = min(group), max(group)
            else:
                child_global = tuple(sorted(group))
                lo = min(mnodes[gi].leaf_lo for gi in group)
                hi = max(mnodes[gi].leaf_hi for gi in group)

            pchunk = _build_parent_chunk_from_leaf_slice(
                leaves, lo, hi, content, filepath
            )
            parent_idx = len(mnodes)
            mnodes.append(
                _MutableNode(
                    chunk=pchunk,
                    level=ell,
                    parent=None,
                    children=list(child_global),
                    leaf_lo=lo,
                    leaf_hi=hi,
                )
            )
            for ci in child_global:
                mnodes[ci].parent = parent_idx
            new_level_indices.append(parent_idx)

        prev_level_indices = new_level_indices

    _finalize_chunk_indices(mnodes)

    nodes = [
        ChunkNode(
            chunk=m.chunk,
            level=m.level,
            parent_index=m.parent,
            child_indices=tuple(sorted(m.children)),
        )
        for m in mnodes
    ]
    return ChunkTree(nodes=nodes, level_count=L)



from __future__ import annotations

import html
import json
import re
import subprocess
import sys
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
HTML_SCRIPT = ROOT / "benchmarks" / "run_html_report.py"


def test_run_html_report_importable_without_running_main() -> None:
    assert "if __name__" in HTML_SCRIPT.read_text(encoding="utf-8")


def test_run_html_report_generates_file(tmp_path: Path) -> None:
    out = tmp_path / "report.html"
    result = subprocess.run(
        [
            sys.executable,
            str(HTML_SCRIPT),
            "--output",
            str(out),
            "--repeat",
            "1",
            "--scenarios",
            "sample_json",
        ],
        cwd=str(ROOT),
        check=False,
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0, result.stderr
    assert out.is_file()
    body = out.read_text(encoding="utf-8")
    assert "omnichunk" in body
    assert "Throughput" in body
    assert "Quality" in body


def test_run_html_report_raw_json_valid(tmp_path: Path) -> None:
    out = tmp_path / "rep.html"
    subprocess.run(
        [
            sys.executable,
            str(HTML_SCRIPT),
            "--output",
            str(out),
            "--repeat",
            "1",
            "--scenarios",
            "sample_json",
        ],
        cwd=str(ROOT),
        check=True,
        capture_output=True,
    )
    body = out.read_text(encoding="utf-8")
    m = re.search(r'<pre id="omnichunk-raw-json">(.*?)</pre>', body, re.DOTALL)
    assert m
    raw = html.unescape(m.group(1).strip())
    data = json.loads(raw)
    assert "throughput" in data
    assert "quality" in data



from __future__ import annotations

from collections.abc import Iterator

from omnichunk.semantic.splitter import SemanticSplitter
from omnichunk.types import Chunk, ChunkOptions


class SemanticEngine:
    """Engine for embedding-based semantic chunking of prose content."""

    def chunk(self, filepath: str, content: str, options: ChunkOptions) -> list[Chunk]:
        splitter = self._build_splitter(options)
        return splitter.split(filepath, content, options)

    def stream(self, filepath: str, content: str, options: ChunkOptions) -> Iterator[Chunk]:
        chunks = self.chunk(filepath, content, options)
        for idx, ch in enumerate(chunks):
            yield _with_unknown_total(ch, idx)

    def _build_splitter(self, options: ChunkOptions) -> SemanticSplitter:
        embed_fn = options.semantic_embed_fn
        if embed_fn is None or not callable(embed_fn):
            raise ValueError(
                "semantic=True requires semantic_embed_fn: "
                "Callable[[list[str]], np.ndarray]"
            )
        ss = options.semantic_sentence_splitter
        sentence_fn = ss if callable(ss) else None
        return SemanticSplitter(
            embed_fn=embed_fn,
            window=int(options.semantic_window),
            threshold=float(options.semantic_threshold),
            min_chunk_sentences=max(1, int(options.semantic_min_sentences)),
            sentence_splitter_fn=sentence_fn,
        )


def _with_unknown_total(chunk: Chunk, index: int) -> Chunk:
    return Chunk(
        text=chunk.text,
        contextualized_text=chunk.contextualized_text,
        byte_range=chunk.byte_range,
        line_range=chunk.line_range,
        index=index,
        total_chunks=-1,
        context=chunk.context,
        token_count=chunk.token_count,
        char_count=chunk.char_count,
        nws_count=chunk.nws_count,
    )
